import itertools
from abc import abstractmethod
from typing import Iterable


def _check_traversal_path_type(traversal_paths):
    if isinstance(traversal_paths, str):
        raise ValueError('traversal_paths needs to be an Iterable[str]')


class TraversableSequence:
    """
    A mixin used for traversing a `Sequence[Traversable]`.
    """

    def traverse(
        self, traversal_paths: Iterable[str]
    ) -> Iterable['TraversableSequence']:
        """
        Return an Iterator of :class:``TraversableSequence`` of the leaves when applying the traversal_paths.
        Each :class:``TraversableSequence`` is either the root Documents, a ChunkArray or a MatchArray.

        :param traversal_paths: a list of string that represents the traversal path
        :yield: :class:``TraversableSequence`` of the leaves when applying the traversal_paths.

        Example on ``traversal_paths``:

            - [`r`]: docs in this TraversableSequence
            - [`m`]: all match-documents at adjacency 1
            - [`c`]: all child-documents at granularity 1
            - [`cc`]: all child-documents at granularity 2
            - [`mm`]: all match-documents at adjacency 2
            - [`cm`]: all match-document at adjacency 1 and granularity 1
            - [`r`, `c`]: docs in this TraversableSequence and all child-documents at granularity 1

        """
        _check_traversal_path_type(traversal_paths)

        for p in traversal_paths:
            yield from self._traverse(self, p)

    @staticmethod
    def _traverse(docs: 'TraversableSequence', path: str):
        if path:
            loc = path[0]
            if loc == 'r':
                yield from TraversableSequence._traverse(docs, path[1:])
            elif loc == 'm':
                for d in docs:
                    yield from TraversableSequence._traverse(d.matches, path[1:])
            elif loc == 'c':
                for d in docs:
                    yield from TraversableSequence._traverse(d.chunks, path[1:])
            else:
                raise ValueError(
                    f'`path`:{loc} is invalid, must be one of `c`, `r`, `m`'
                )
        else:
            yield docs

    def traverse_flat_per_path(
        self, traversal_paths: Iterable[str]
    ) -> Iterable['TraversableSequence']:
        """
        Returns a flattened :class:``TraversableSequence`` per path in :param:``traversal_paths``
        with all Documents, that are reached by the path.

        :param traversal_paths: a list of string that represents the traversal path
        :yield: :class:``TraversableSequence`` containing the document of all leaves per path.
        """
        _check_traversal_path_type(traversal_paths)

        for p in traversal_paths:
            yield self._flatten(self._traverse(self, p))

    def traverse_flat(self, traversal_paths: Iterable[str]) -> 'TraversableSequence':
        """
        Returns a single flattened :class:``TraversableSequence`` with all Documents, that are reached
        via the :param:``traversal_paths``.

        .. warning::
            When defining the :param:``traversal_paths`` with multiple paths, the returned
            :class:``Documents`` are determined at once and not on the fly. This is a different
            behavior then in :method:``traverse`` and :method:``traverse_flattened_per_path``!

        :param traversal_paths: a list of string that represents the traversal path
        :return: a single :class:``TraversableSequence`` containing the document of all leaves when applying the traversal_paths.
        """
        _check_traversal_path_type(traversal_paths)

        leaves = self.traverse(traversal_paths)
        return self._flatten(leaves)

    @staticmethod
    @abstractmethod
    def _flatten(sequence):
        ...



from typing import List

import numpy as np
import torch
from torch.sparse import FloatTensor

from . import BaseSparseNdArray
from ....proto import jina_pb2

if False:
    import numpy as np

__all__ = ['SparseNdArray']


class SparseNdArray(BaseSparseNdArray):
    """
    Pytorch powered sparse ndarray, i.e. FloatTensor.

    .. seealso::
        https://pytorch.org/docs/stable/sparse.html

    :param proto: protobuf instance, default is None.
    :param transpose_indices: in torch, the input to LongTensor is NOT a list of index tuples.
    :param args: positional arguments.
    :param kwargs: positional key value arguments.

    If you want to write your indices this way, you should transpose before passing them to the sparse constructor

    .. note::
        To comply with Tensorflow, `transpose_indices` is set to True by default
    """

    def __init__(
        self,
        proto: 'jina_pb2.SparseNdArrayProto' = None,
        transpose_indices: bool = True,
        *args,
        **kwargs
    ):
        super().__init__(proto, *args, **kwargs)
        self.transpose_indices = transpose_indices

    def sparse_constructor(
        self, indices: 'np.ndarray', values: 'np.ndarray', shape: List[int]
    ) -> 'FloatTensor':
        """
        Sparse NdArray constructor for FloatTensor.

        :param indices: the indices of the sparse array
        :param values: the values of the sparse array
        :param shape: the shape of the sparse array
        :return: FloatTensor
        """
        return FloatTensor(
            torch.LongTensor(indices).T
            if self.transpose_indices
            else torch.LongTensor(indices),
            torch.FloatTensor(values),
            torch.Size(shape),
        )

    def sparse_parser(self, value: 'FloatTensor'):
        """
        Parse a FloatTensor to indices, values and shape.

        :param value: the FloatTensor.
        :return: a Dict with three entries {'indices': ..., 'values':..., 'shape':...}
        """
        indices = value._indices().numpy()
        if self.transpose_indices:
            indices = indices.T
        values = value._values().numpy()
        shape = list(value.size())
        return {
            'indices': indices,
            'values': values,
            'shape': shape,
        }



import multiprocessing
import threading
from functools import partial
from typing import Union, TYPE_CHECKING
from copy import deepcopy

from ... import __default_host__
from ...hubble.hubio import HubIO
from ...hubble.helper import is_valid_huburi
from ...enums import GatewayProtocolType, RuntimeBackendType
from ...parsers.hubble import set_hub_pull_parser

if TYPE_CHECKING:
    from argparse import Namespace


def _get_event(obj) -> Union[multiprocessing.Event, threading.Event]:
    if isinstance(obj, threading.Thread):
        return threading.Event()
    elif isinstance(obj, multiprocessing.Process) or isinstance(
        obj, multiprocessing.context.ForkProcess
    ):
        return multiprocessing.Event()
    elif isinstance(obj, multiprocessing.context.SpawnProcess):
        return multiprocessing.get_context('spawn').Event()
    else:
        raise TypeError(
            f'{obj} is not an instance of "threading.Thread" nor "multiprocessing.Process"'
        )


class ConditionalEvent:
    """
    :class:`ConditionalEvent` provides a common interface to an event (multiprocessing or threading event)
    that gets triggered when any of the events provided in input is triggered (OR logic)

    :param backend_runtime: The runtime type to decide which type of Event to instantiate
    :param events_list: The list of events that compose this composable event
    """

    def __init__(self, backend_runtime: RuntimeBackendType, events_list):
        super().__init__()
        self.event = None
        if backend_runtime == RuntimeBackendType.THREAD:
            self.event = threading.Event()
        else:
            self.event = multiprocessing.synchronize.Event(
                ctx=multiprocessing.get_context()
            )
        self.event_list = events_list
        for e in events_list:
            self._setup(e, self._state_changed)

        self._state_changed()

    def _state_changed(self):
        bools = [e.is_set() for e in self.event_list]
        if any(bools):
            self.event.set()
        else:
            self.event.clear()

    def _custom_set(self, e):
        e._set()
        e._state_changed()

    def _custom_clear(self, e):
        e._clear()
        e._state_changed()

    def _setup(self, e, changed_callback):
        e._set = e.set
        e._clear = e.clear
        e._state_changed = changed_callback
        e.set = partial(self._custom_set, e)
        e.clear = partial(self._custom_clear, e)


def update_runtime_cls(args, copy=False) -> 'Namespace':
    """Get runtime_cls as a string from args

    :param args: pea/pod namespace args
    :param copy: True if args shouldn't be modified in-place
    :return: runtime class as a string
    """
    _args = deepcopy(args) if copy else args
    gateway_runtime_dict = {
        GatewayProtocolType.GRPC: 'GRPCRuntime',
        GatewayProtocolType.WEBSOCKET: 'WebSocketRuntime',
        GatewayProtocolType.HTTP: 'HTTPRuntime',
    }
    if (
        _args.runtime_cls not in gateway_runtime_dict.values()
        and _args.host != __default_host__
        and not _args.disable_remote
    ):
        _args.runtime_cls = 'JinadRuntime'
        # NOTE: remote pea would also create a remote workspace which might take alot of time.
        # setting it to -1 so that wait_start_success doesn't fail
        _args.timeout_ready = -1
    if _args.runtime_cls == 'ZEDRuntime' and _args.uses.startswith('docker://'):
        _args.runtime_cls = 'ContainerRuntime'
    if _args.runtime_cls == 'ZEDRuntime' and is_valid_huburi(_args.uses):
        _hub_args = [_args.uses, '--no-usage']
        if _args.install_requirements:
            _hub_args.append('--install-requirements')

        _args.uses = HubIO(set_hub_pull_parser().parse_args(_hub_args)).pull()
        if _args.uses.startswith('docker://'):
            _args.runtime_cls = 'ContainerRuntime'

    if hasattr(_args, 'protocol'):
        _args.runtime_cls = gateway_runtime_dict[_args.protocol]

    return _args
